"""
Platform-specific utilities such as library imports and external API
definitions.

.. note::

    The individual files are typically not included in documentation as
    they have import side effects and are platform specific.
"""
